from collections import deque
from typing import Any, Iterable, Union


class Queue:
    def __init__(self):
        """初始化队列"""
        self._queue = deque()

    @property
    def length(self) -> int:
        """返回队列长度"""
        return len(self._queue)

    def empty(self) -> bool:
        """检查队列是否为空"""
        return not self._queue

    def push(self, data: Union[Any, Iterable], direct_insertion: bool = False):
        """
        向队列中添加数据。
        如果数据是可迭代对象（如列表、元组或集合），可以选择直接扩展队列。
        """
        if not data:
            return

        if not direct_insertion and isinstance(data, (list, tuple, set)):
            self._queue.extend(data)
        else:
            self._queue.append(data)

    def pop(self) -> Any:
        """
        从队列左侧弹出一个元素。
        如果队列为空，返回 None。
        """
        try:
            return self._queue.popleft()
        except IndexError:
            return None

    def iter_items(self, limit: int = 1) -> Iterable:
        """
        按指定数量从队列中弹出多个元素并生成它们。
        如果队列为空或达到限制，则停止生成。
        """
        for _ in range(limit):
            item = self.pop()
            if item is None:
                break
            yield item


import os
import time
import threading
from typing import Callable
from cobweb.db import RedisDB, ApiDB
from cobweb.utils import check_pause
from cobweb.base import Queue, Seed, logger
from cobweb.constant import LogTemplate
from .scheduler import Scheduler
use_api = bool(os.getenv("REDIS_API_HOST", 0))


class RedisScheduler(Scheduler):

    def __init__(
            self,
            task,
            project,
            stop: threading.Event,
            pause: threading.Event,
            new: Queue,
            todo: Queue,
            done: Queue,
            upload: Queue,
            callback_register: Callable
    ):
        super().__init__(task, project, stop, pause, new, todo, done, upload, callback_register)
        self.todo_key = f"{{{project}:{task}}}:todo"
        self.done_key = f"{{{project}:{task}}}:done"
        self.fail_key = f"{{{project}:{task}}}:fail"
        self.heartbeat_key = f"heartbeat:{project}_{task}"
        self.speed_control_key = f"speed_control:{project}_{task}"
        self.reset_lock_key = f"lock:reset:{project}_{task}"
        self.db = ApiDB() if use_api else RedisDB()

    def reset(self):
        """
        检查过期种子，重新添加到redis缓存中
        """
        while not self.stop.is_set():
            if self.db.lock(self.reset_lock_key, t=60):

                _min = -int(time.time()) + self.seed_reset_seconds
                self.db.members(self.todo_key, 0, _min=_min, _max="(0")
                self.db.delete(self.reset_lock_key)

            time.sleep(60)

    @check_pause
    def schedule(self):
        """
        调度任务，获取redis队列种子，同时添加到doing字典中
        """
        if not self.db.zcount(self.todo_key, 0, "(1000"):
            time.sleep(self.scheduler_wait_seconds)
            return

        if self.todo.length >= self.todo_queue_size:
            time.sleep(self.todo_queue_full_wait_seconds)
            return

        count = self.get_working_items_count()
        if count > 500:
            time.sleep(self.todo_queue_full_wait_seconds)
            logger.info(f"任务挤压, 内存任务数：{count}")
            return

        members = self.db.members(
            self.todo_key, int(time.time()),
            count=self.todo_queue_size,
            _min=0, _max="(1000"
        )

        logger.debug(f"Retrieved {len(members)} seeds from Redis.")

        seeds, item_info = list(), dict()
        for member, priority in members:
            seed = Seed(member, priority=int(priority % 1000))
            item_info[seed.to_string] = seed.params.priority
            seeds.append(seed)

        self.set_working_items(item_info)
        self.todo.push(seeds)
        time.sleep(1)

    @check_pause
    def insert(self):
        """
        添加新种子到redis队列中
        """
        seeds, delete_seeds = dict(), set()
        for seed, new_seed in self.new.iter_items(limit=self.new_queue_max_size):
            seeds[new_seed.to_string] = new_seed.params.priority
            delete_seeds.add(seed)

        self.db.zadd(self.todo_key, seeds, nx=True)
        self.done.push(delete_seeds)

        if self.new.length < self.new_queue_max_size:
            time.sleep(self.scheduler_wait_seconds)

    @check_pause
    def refresh(self):
        """
        刷新doing种子过期时间，防止reset重新消费
        """
        if item_info := self.get_working_items():
            refresh_time = int(time.time())
            seed_info = {k: -refresh_time - v / 1000 for k, v in item_info.items()}
            self.db.zadd(self.todo_key, seed_info, xx=True)
            # self.set_working_items(seed_info)
        time.sleep(20)

    @check_pause
    def delete(self):
        """
        删除队列种子，根据状态添加至成功或失败队列，移除doing字典种子索引
        """
        seeds = [seed for seed in self.done.iter_items(limit=self.done_queue_max_size)]
        items = [seed.to_string for seed in seeds]

        if self.remove_working_items(items):
            self.db.zrem(self.todo_key, *items)
        elif seeds:
            self.done.push(seeds)

        if self.done.length < self.done_queue_max_size:
            time.sleep(self.done_queue_wait_seconds)

    def run(self):
        start_time = int(time.time())

        for func in [self.reset, self.insert, self.delete, self.refresh, self.schedule]:
            self.callback_register(func, tag="scheduler")

        while not self.stop.is_set():
            working_count = self.get_working_items_count()
            memory_count = self.db.zcount(self.todo_key, "-inf", "(0")
            todo_count = self.db.zcount(self.todo_key, 0, "(1000")
            all_count = self.db.zcard(self.todo_key)

            if self.pause.is_set():
                execute_time = int(time.time()) - start_time
                if not self.task_model and execute_time > self.before_scheduler_wait_seconds:
                    logger.info("Done! ready to close thread...")
                    self.stop.set()
                elif todo_count:
                    logger.info(
                        f"Recovery {self.task} task run！Todo seeds count: {todo_count}, queue length: {all_count}")
                    self.pause.clear()
                else:
                    logger.info("Pause! waiting for resume...")

            elif self.is_empty():

                if all_count:
                    logger.info(f"Todo seeds count: {todo_count}, queue length: {all_count}")
                    self.pause.clear()
                else:
                    count = 0
                    for _ in range(3):
                        if not all_count:
                            count += 1
                            time.sleep(5)
                            logger.info("Checking count...")
                        else:
                            break
                    if count >= 3:
                        logger.info("Todo queue is empty! Pause set...")
                        self.clear_working_items()
                        self.pause.set()

            else:
                logger.info(LogTemplate.launcher_pro_polling.format(
                    task=self.task,
                    doing_len=working_count,
                    todo_len=self.todo.length,
                    done_len=self.done.length,
                    redis_seed_count=all_count,
                    redis_todo_len=todo_count,
                    redis_doing_len=memory_count,
                    upload_len=self.upload.length,
                ))

            time.sleep(30)



import random
import requests
from typing import Any, Dict


class Request:
    """
    请求类，用于封装 HTTP 请求并提供相关功能。
    """

    __REQUEST_ATTRS__ = {
        "params",
        "headers",
        "cookies",
        "data",
        "json",
        "files",
        "auth",
        "timeout",
        "proxies",
        "hooks",
        "stream",
        "verify",
        "cert",
        "allow_redirects",
    }

    def __init__(
        self,
        url: str,
        seed: Any,
        random_ua: bool = True,
        check_status_code: bool = True,
        **kwargs,
    ):
        """
        初始化请求对象。
        :param url: 请求的 URL。
        :param seed: 种子对象或标识符。
        :param random_ua: 是否使用随机 User-Agent，默认为 True。
        :param check_status_code: 是否检查响应状态码，默认为 True。
        :param kwargs: 其他扩展参数。
        """
        self.url = url
        self.seed = seed
        self.check_status_code = check_status_code
        self.request_setting: Dict[str, Any] = {}

        for key, value in kwargs.items():
            if key in self.__class__.__REQUEST_ATTRS__:
                self.request_setting[key] = value
            else:
                setattr(self, key, value)

        self.method = getattr(self, "method", None) or (
            "POST" if self.request_setting.get("data") or self.request_setting.get("json") else "GET"
        )

        if random_ua:
            self._build_header()

    @property
    def _random_ua(self) -> str:
        v1 = random.randint(4, 15)
        v2 = random.randint(3, 11)
        v3 = random.randint(1, 16)
        v4 = random.randint(533, 605)
        v5 = random.randint(1000, 6000)
        v6 = random.randint(10, 80)
        return (f"Mozilla/5.0 (Macintosh; Intel Mac OS X 10_{v1}_{v2}) "
                f"AppleWebKit/{v4}.{v3} (KHTML, like Gecko) "
                f"Chrome/105.0.0.0 Safari/{v4}.{v3} Edg/105.0.{v5}.{v6}")

    def _build_header(self):
        headers = self.request_setting.setdefault("headers", {})
        if not headers.get("user-agent"):
            headers["user-agent"] = self._random_ua

    def download(self) -> requests.Response:
        response = requests.request(self.method, self.url, **self.request_setting)
        if self.check_status_code:
            response.raise_for_status()
        return response

    @property
    def to_dict(self) -> Dict[str, Any]:
        excluded_keys = {"url", "seed", "check_status_code", "request_setting"}
        return {k: v for k, v in self.__dict__.items() if k not in excluded_keys}


import time
import threading
import traceback

from typing import Callable, Type
from inspect import isgenerator
from urllib.parse import urlparse
from requests import Response as Res

from cobweb.crawlers import Crawler
from cobweb.constant import DealModel, LogTemplate
from cobweb.utils import LoghubDot, check_pause
from cobweb.base import Seed, Queue, BaseItem, Request, Response, logger


class Distributor(threading.Thread):

    def __init__(
            self,
            task: str,
            project: str,
            new: Queue,
            todo: Queue,
            done: Queue,
            upload: Queue,
            stop: threading.Event,
            pause: threading.Event,
            callback_register: Callable,
            SpiderCrawler: Type[Crawler]
    ):
        super().__init__()
        self.task = task
        self.project = project
        self.stop = stop
        self.pause = pause

        self.new = new
        self.todo = todo
        self.done = done
        self.upload = upload
        self.callback_register = callback_register
        self.Crawler = SpiderCrawler

        from cobweb import setting
        self.time_sleep = setting.SPIDER_TIME_SLEEP
        self.thread_num = setting.SPIDER_THREAD_NUM
        self.max_retries = setting.SPIDER_MAX_RETRIES
        self.record_failed = setting.RECORD_FAILED_SPIDER
        self.loghub_dot = LoghubDot(stop=stop)  # todo: 解偶

        logger.debug(f"Distribute instance attrs: {self.__dict__}")

    def distribute(self, item, seed, _id: int):
        if isinstance(item, Request):
            seed.params.start_time = time.time()
            self.process(item=item, seed=seed, callback=self.Crawler.download, _id=1)
        elif isinstance(item, Response):
            if _id == 2:
                raise TypeError("parse function can't yield a Response instance")
            dot = isinstance(item.response, Res)
            # TODO: 请求成功打点
            self.spider_logging(seed, item, dot=dot)
            self.process(item=item, seed=seed, callback=self.Crawler.parse, _id=2)
        elif isinstance(item, BaseItem):
            self.upload.push(item)
        elif isinstance(item, Seed):
            self.new.push((seed, item), direct_insertion=True)
        elif isinstance(item, str) and item == DealModel.poll:
            self.todo.push(seed)
        elif isinstance(item, str) and item == DealModel.done:
            self.done.push(seed)
        elif isinstance(item, str) and item == DealModel.fail:
            seed.params.retry += 1
            if seed.params.retry < self.max_retries:
                self.todo.push(seed)
            else:
                if record_failed := self.record_failed:
                    try:
                        response = Response(seed, "failed", max_retries=True)
                        self.process(response, seed, self.Crawler.parse, _id=2)
                    except Exception as e:
                        msg = ''.join(traceback.format_exception(type(e), e, e.__traceback__))
                        logger.error(msg = msg)
                        record_failed = False
                if not record_failed:
                    self.done.push(seed)
        else:
            raise TypeError("yield value type error!")

    def process(self, item, seed, callback, _id: int):
        result_iterators = callback(item)
        if not isgenerator(result_iterators):
            raise TypeError(f"{callback.__name__} function isn't a generator!")
        for result_item in result_iterators:
            self.distribute(result_item, seed, _id)

    @check_pause
    def spider(self):
        # TODO: 限流措施
        if seed := self.todo.pop():
            try:
                self.process(item=seed, seed=seed, callback=self.Crawler.request, _id=0)
            except Exception as e:
                url, status = seed.url, e.__class__.__name__
                msg = ''.join(traceback.format_exception(type(e), e, e.__traceback__))
                if getattr(e, "response", None) and isinstance(e.response, Res):
                    url = e.response.request.url
                    status = e.response.status_code
                    # TODO:失败请求打点
                self.spider_logging(seed, None, error=True, url=url, status=status, msg=msg)
                self.distribute(DealModel.fail, seed, _id=-1)

    def spider_logging(
            self, seed,
            item: Response = None,
            error: bool = False,
            dot: bool = True,
            **kwargs
    ):
        detail_log_info = LogTemplate.log_info(seed.to_dict)
        if error:
            url = kwargs.get("url")
            msg = kwargs.get("msg")
            status = kwargs.get("status")
            if dot:
                self.loghub_dot.build(
                    topic=urlparse(url).netloc,
                    data_size=-1, cost_time=-1,
                    status=status, url=url,
                    seed=seed.to_string,
                    proxy_type=seed.params.proxy_type,
                    proxy=seed.params.proxy,
                    project=self.project,
                    task=self.task, msg=msg,
                )
            logger.info(LogTemplate.download_exception.format(
                detail=detail_log_info,
                retry=seed.params.retry,
                priority=seed.params.priority,
                seed_version=seed.params.seed_version,
                identifier=seed.identifier or "",
                exception=msg
            ))
        else:
            logger.info(LogTemplate.download_info.format(
                detail=detail_log_info,
                retry=seed.params.retry,
                priority=seed.params.priority,
                seed_version=seed.params.seed_version,
                identifier=seed.identifier or "",
                status=item.response,
                response=LogTemplate.log_info(item.to_dict)
            ))
            if dot:
                end_time = time.time()
                stime = seed.params.start_time
                cost_time = end_time - stime if stime else -1
                topic = urlparse(item.response.request.url).netloc
                data_size = int(item.response.headers.get("content-length", 0))
                self.loghub_dot.build(
                    topic=topic, data_size=data_size, cost_time=cost_time,
                    status=200, seed=seed.to_string, url=item.response.url,
                    proxy=seed.params.proxy, proxy_type=seed.params.proxy_type,
                    project=self.project, task=self.task,
                )

    def run(self):
        self.callback_register(self.loghub_dot.build_run, tag="LoghubDot")
        for _ in range(self.thread_num):
            self.callback_register(self.spider, tag="Distributor")



from abc import ABC, abstractmethod
from cobweb.base import BaseItem


class Pipeline(ABC):

    def __init__(self, task, project):
        self.task = task
        self.project = project

    @abstractmethod
    def build(self, item: BaseItem) -> dict:
        pass

    @abstractmethod
    def upload(self, table: str, data: list) -> bool:
        pass





import threading


from typing import Callable
from cobweb.base import Queue
from abc import ABC, abstractmethod


class Scheduler(ABC, threading.Thread):

    __WORKING_ITEMS__ = {}
    __LAUNCHER_FUNC__ = ["_reset", "_scheduler", "_insert", "_refresh", "_delete"]

    def __init__(
            self,
            task,
            project,
            stop: threading.Event,
            pause: threading.Event,
            new: Queue,
            todo: Queue,
            done: Queue,
            upload: Queue,
            callback_register: Callable
    ):
        super().__init__()
        self.task = task
        self.project = project
        from cobweb import setting

        self.task_model = setting.TASK_MODEL
        self.seed_reset_seconds = setting.SEED_RESET_SECONDS
        self.scheduler_wait_seconds = setting.SCHEDULER_WAIT_SECONDS
        self.new_queue_wait_seconds = setting.NEW_QUEUE_WAIT_SECONDS
        self.done_queue_wait_seconds = setting.DONE_QUEUE_WAIT_SECONDS
        self.todo_queue_full_wait_seconds = setting.TODO_QUEUE_FULL_WAIT_SECONDS
        self.before_scheduler_wait_seconds = setting.BEFORE_SCHEDULER_WAIT_SECONDS

        self.todo_queue_size = setting.TODO_QUEUE_SIZE
        self.new_queue_max_size = setting.NEW_QUEUE_MAX_SIZE
        self.done_queue_max_size = setting.DONE_QUEUE_MAX_SIZE
        self.upload_queue_max_size = setting.UPLOAD_QUEUE_MAX_SIZE

        self.stop = stop
        self.pause = pause

        self.new = new
        self.todo = todo
        self.done = done
        self.upload = upload

        self.callback_register = callback_register
        self.lock = threading.Lock()

    def is_empty(self):
        return all(queue.empty() for queue in (self.new, self.todo, self.done, self.upload))

    def set_working_items(self, item_info: dict = None):
        if not item_info:
            return False
        with self.lock:
            self.__WORKING_ITEMS__.update(item_info)

    def get_working_items(self) -> dict:
        with self.lock:
            return self.__WORKING_ITEMS__.copy()

    def remove_working_items(self, items: list[str] = None) -> int:
        if not items:
            return 0
        with self.lock:
            deleted_count = 0
            for item in items:
                if item in self.__WORKING_ITEMS__:
                    del self.__WORKING_ITEMS__[item]
                    deleted_count += 1
            return deleted_count

    def get_working_items_count(self) -> int:
        with self.lock:
            return len(self.__WORKING_ITEMS__)

    def clear_working_items(self):
        with self.lock:
            self.__WORKING_ITEMS__.clear()

    @abstractmethod
    def reset(self):
        ...

    @abstractmethod
    def schedule(self):
        ...

    @abstractmethod
    def insert(self):
        ...

    @abstractmethod
    def refresh(self):
        ...

    @abstractmethod
    def delete(self):
        ...





import re
import hashlib
import inspect
from typing import Union
from importlib import import_module


def md5(text: Union[str, bytes]) -> str:
    if isinstance(text, str):
        text = text.encode('utf-8')
    return hashlib.md5(text).hexdigest()


def dynamic_load_class(model_info):
    if isinstance(model_info, str):
        if "import" in model_info:
            model_path, class_name = re.search(
                r"from (.*?) import (.*?)$", model_info
            ).groups()
            model = import_module(model_path)
            class_object = getattr(model, class_name)
        else:
            model_path, class_name = model_info.rsplit(".", 1)
            model = import_module(model_path)
            class_object = getattr(model, class_name)
        return class_object
    elif inspect.isclass(model_info):
        return model_info
    raise TypeError()




import time
import uuid
import inspect
import threading
import importlib

from cobweb import setting
from cobweb.base import Seed, Queue, logger
from cobweb.utils.tools import dynamic_load_class
from cobweb.launchers.distributor import Distributor
from cobweb.launchers.uploader import Uploader
from typing import Optional, Union, Dict, Any, Callable


class Launcher:

    _NEW_QUEUE_ = Queue()
    _TODO_QUEUE_ = Queue()
    _DONE_QUEUE_ = Queue()
    _UPLOAD_QUEUE_ = Queue()

    __REGISTER_FUNC__: Dict[str, Callable] = {}
    __WORKER_THREAD__: Dict[str, threading.Thread] = {}

    def __init__(self, task, project, custom_setting=None, **kwargs):
        super().__init__()

        self.task = task
        self.project = project

        self._app_time = int(time.time())
        self._stop = threading.Event()  # 结束事件
        self._pause = threading.Event()  # 暂停事件

        _setting = self._load_custom_settings(custom_setting)
        _setting.update(kwargs)
        for key, value in _setting.items():
            setattr(setting, key.upper(), value)

        self._done_model = setting.DONE_MODEL
        self._task_model = setting.TASK_MODEL

        self.Scheduler = dynamic_load_class(setting.SCHEDULER)
        self.SpiderCrawler = dynamic_load_class(setting.CRAWLER)
        self.SpiderPipeline = dynamic_load_class(setting.PIPELINE)

    @staticmethod
    def _load_custom_settings(custom_setting: Optional[Union[str, Dict]]) -> Dict[str, Any]:
        _setting = {}
        if custom_setting:
            if isinstance(custom_setting, dict):
                _setting = custom_setting
            elif isinstance(custom_setting, str):
                module = importlib.import_module(custom_setting)
                _setting = {
                    k: v
                    for k, v in module.__dict__.items()
                    if not k.startswith("__") and not inspect.ismodule(v)
                }
            else:
                raise ValueError("custom_setting must be a dictionary or a module path.")
        return _setting

    @property
    def request(self):
        """
        自定义request函数
        use case:
            from cobweb.base import Request, BaseItem
            @launcher.request
            def request(seed: Seed) -> Union[Request, BaseItem]:
                ...
                yield Request(seed.url, seed)
        """
        def decorator(func):
            self.SpiderCrawler.request = func
        return decorator

    @property
    def download(self):
        """
        自定义download函数
        use case:
            from cobweb.base import Request, Response, Seed, BaseItem
            @launcher.download
            def download(item: Request) -> Union[Seed, BaseItem, Response, str]:
                ...
                yield Response(item.seed, response)
        """
        def decorator(func):
            self.SpiderCrawler.download = func
        return decorator

    @property
    def parse(self):
        """
        自定义parse函数, xxxItem为自定义的存储数据类型
        use case:
            from cobweb.base import Request, Response
            @launcher.parse
            def parse(item: Response) -> BaseItem:
               ...
               yield xxxItem(seed, **kwargs)
        """
        def decorator(func):
            self.SpiderCrawler.parse = func
        return decorator

    def start_seeds(self, seeds: list[Union[str, Dict]]) -> list[Seed]:
        seed_list = [Seed(seed) for seed in seeds]
        self._TODO_QUEUE_.push(seed_list)
        return seed_list

    def _register(self, func: Callable, tag: str = "launcher"):
        name = f"{tag}:{func.__name__}_{uuid.uuid4()}"
        self.__REGISTER_FUNC__[name] = func
        if not self.__WORKER_THREAD__.get(name):
            worker_thread = threading.Thread(name=name, target=func)
            self.__WORKER_THREAD__[name] = worker_thread

    def _monitor(self):
        while not self._stop.is_set():
            if not self._pause.is_set():
                for name, worker_thread in list(self.__WORKER_THREAD__.items()):
                    if not worker_thread.is_alive():
                        logger.debug(f"{name} thread is dead. Restarting...")
                        func = self.__REGISTER_FUNC__[name]
                        worker_thread = threading.Thread(name=name, target=func)
                        self.__WORKER_THREAD__[name] = worker_thread
                        worker_thread.start()
            time.sleep(15)
        logger.info("monitor thread close!")

    def start(self):
        self._pause.is_set()

        self.Scheduler(
            task=self.task,
            project=self.project,
            stop=self._stop,
            pause=self._pause,
            new=self._NEW_QUEUE_,
            todo=self._TODO_QUEUE_,
            done=self._DONE_QUEUE_,
            upload=self._UPLOAD_QUEUE_,
            callback_register=self._register
        ).start()

        Distributor(
            task=self.task,
            project=self.project,
            new=self._NEW_QUEUE_,
            todo=self._TODO_QUEUE_,
            done=self._DONE_QUEUE_,
            upload=self._UPLOAD_QUEUE_,
            callback_register=self._register,
            stop=self._stop, pause=self._pause,
            SpiderCrawler=self.SpiderCrawler
        ).start()

        Uploader(
            task=self.task, project=self.project,
            stop=self._stop, pause=self._pause,
            done=self._DONE_QUEUE_,
            upload=self._UPLOAD_QUEUE_,
            callback_register=self._register,
            SpiderPipeline=self.SpiderPipeline
        ).start()

        self._monitor()
        logger.info("task done!")




import time
import threading
from typing import Callable, Type
from cobweb.pipelines import Pipeline
from cobweb.base import Queue, logger
from cobweb.utils import check_pause


class Uploader(threading.Thread):

    def __init__(
            self,
            task: str,
            project: str,
            stop: threading.Event,
            pause: threading.Event,
            upload: Queue, done: Queue,
            callback_register: Callable,
            SpiderPipeline: Type[Pipeline]
    ):
        super().__init__()
        self.task = task
        self.project = project

        self.stop = stop
        self.pause = pause

        self.done = done
        self.upload = upload
        self.callback_register = callback_register

        from cobweb import setting

        self.upload_size = setting.UPLOAD_QUEUE_MAX_SIZE
        self.wait_seconds = setting.UPLOAD_QUEUE_WAIT_SECONDS

        self.pipeline = SpiderPipeline(task=self.task, project=self.project)

        logger.debug(f"Uploader instance attrs: {self.__dict__}")

    @check_pause
    def upload_data(self):
        if not self.upload.length:
            time.sleep(self.wait_seconds)
            return
        if self.upload.length < self.upload_size:
            time.sleep(self.wait_seconds)
        data_info, seeds = {}, []
        try:
            for _ in range(self.upload_size):
                item = self.upload.pop()
                if not item:
                    break
                seeds.append(item.seed)
                data = self.pipeline.build(item)
                data_info.setdefault(item.table, []).append(data)
            for table, datas in data_info.items():
                try:
                    self.pipeline.upload(table, datas)
                    # TODO: 上传打点
                except Exception as e:
                    logger.info(e)
        except Exception as e:
            logger.info(e)
        if seeds:
            self.done.push(seeds)

    def run(self):
        self.callback_register(self.upload_data, tag="Uploader")





from typing import Any, Generator
from cobweb.base import (
    Seed,
    BaseItem,
    Request,
    Response,
    CSVItem,
)


class Crawler:

    @staticmethod
    def request(seed: Seed) -> Generator[Request, Response, None]:
        yield Request(seed.url, seed, timeout=5)

    @staticmethod
    def download(item: Request) -> Generator[Response, Any, None]:
        response = item.download()
        yield Response(item.seed, response, **item.to_dict)

    @staticmethod
    def parse(item: Response) -> Generator[BaseItem, Any, None]:
        upload_item = item.to_dict
        upload_item["content"] = getattr(item.response, "text", item.response)
        yield CSVItem(item.seed, data=upload_item)




import os
import csv

from cobweb.base import BaseItem
from cobweb.pipelines import Pipeline


class CSV(Pipeline):

    def __init__(self, *args, **kwargs):
        super(CSV, self).__init__(*args, **kwargs)
        self.log_path = rf"{os.getcwd()}\{self.project}\{self.task}\%s.csv"

    def build(self, item: BaseItem):
        return item.to_dict

    def upload(self, table, datas):
        fieldnames = datas[0].keys()
        file_path = self.log_path % table
        os.makedirs(os.path.dirname(file_path), exist_ok=True)
        with open(file_path, mode='a', encoding='utf-8', newline="") as file:
            writer = csv.DictWriter(file, fieldnames=fieldnames)
            if file.tell() == 0:  # 判断文件是否为空
                writer.writeheader()
            writer.writerows(datas)



from setuptools import setup, find_packages

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(
    name="cobweb-launcher",
    version="3.1.21",
    packages=find_packages(),
    url="https://github.com/Juannie-PP/cobweb",
    license="MIT",
    author="Juannie-PP",
    author_email="2604868278@qq.com",
    description="spider_hole",
    long_description=long_description,
    long_description_content_type="text/markdown",
    install_requires=["requests>=2.19.1", "redis>=4.4.4", "aliyun-log-python-sdk"],
    classifiers=[
        "Programming Language :: Python :: 3",
    ],
    keywords=[
        "cobweb-launcher, cobweb",
    ],
    python_requires=">=3.7",
)



from .seed import Seed
from typing import Dict, Any
from collections import namedtuple


class ItemMeta(type):

    def __new__(cls, name: str, bases: tuple, dct: dict) -> type:
        new_class = super().__new__(cls, name, bases, dct)
        if name != "BaseItem":
            table = getattr(new_class, "__TABLE__")
            fields = getattr(new_class, "__FIELDS__")
            if not table or not fields:
                raise ValueError(f"Missing required attributes '__TABLE__' or '__FIELDS__' in class {name}")
            new_class.Data = namedtuple(table, fields)
        return new_class


class BaseItem(metaclass=ItemMeta):

    __TABLE__ = ""
    __FIELDS__ = ""

    def __init__(self, seed: Seed, **kwargs):
        self.seed = seed

        data = {}
        for key, value in kwargs.items():
            if key in self.__FIELDS__:
                data[key] = value
            else:
                setattr(self, key, value)

        try:
            self.data = self.Data(**data)
        except TypeError as e:
            raise ValueError(f"Invalid field values for Data: {e}") from e

    @property
    def to_dict(self) -> Dict[str, Any]:
        return self.data._asdict()

    @property
    def fields(self) -> tuple[str]:
        return self.data._fields

    @property
    def table(self) -> str:
        return self.Data.__name__

    def __setitem__(self, key: str, value: Any):
        setattr(self, key, value)

    def __getitem__(self, key: str) -> Any:
        return getattr(self, key, None)

    def __getattr__(self, name: str) -> Any:
        return None


class CSVItem(BaseItem):

    __TABLE__ = "cobweb"
    __FIELDS__ = "data"

